import { ImageListItem, ListOfImages } from '../lib/ListOfImages'
import { BaseSpot, DisplaySpot, MobileSpot, Spot } from '../system/Spot'
import { callable } from '../system_lib/Metadata'
import { Script, ScriptEnv } from '../system_lib/Script'
import { IndexedProperty } from '../system_lib/ScriptBase'

/**
 * What we need to know about our image source (implemented by DisplaySpot and MobileSpot,
 * but TS compiler seemed to need a little extra hint here to accept this type intersection.
 */
interface ImageProvider {
	subscribe(event: 'image', listener: (sender: BaseSpot, message: {
		readonly filePath: string,	// Path to file just received (typically "/temp/xxx/xxx.jpeg")
		readonly rollName: string	// Camera Block's assigned "roll name"
	})=>void): void;

	subscribe(event: 'finish', listener: (sender: BaseSpot)=>void): void;
}

export class PhotoManager extends Script {

	/**
	 * A list of spots paths to spots that can provide images.
	 *
	 * The spots must either be of type DisplaySpot or MobileSpot (visitor spot).
	 */
	private static readonly CAMERA_SPOTS = [
		'Visitor',
		'Station',
		// 'GROUP.SPOT_NAME' // And so on...
	]

	private listOfImages: ListOfImages
	private photoStream: IndexedProperty<ImageListItem>

	constructor(env: ScriptEnv) {
		super(env);

		/**
		 * Create an instance of ListOfImages, which will manage our list
		 * of images. It handles persistence, moving images to the public folder,
		 * and resizing large images etc.
		 */
		this.listOfImages = new ListOfImages(
			this,
			'Photos',
			{
				maxImageSideLength: 1920,
				maxImageCount: 12
			}
		);

		/**
		 * Expose the IndexedProperty created by listOfImages to be able to bind it to the UI elements.
		 */
		this.photoStream = this.listOfImages.getIndexedProperty('photoStream');

		/**
		 * Start listening for photos being taken on the spots we're interested in.
		 */
		this.listenForPhotos()
	}

	/**
	 * For every spot in CAMERA_SPOTS, subscribe to the "image" message, which is sent when a new image is taken,
	 * pass the temporary file path to listOfImages in order to resize, move and persist it.
	 */
	private listenForPhotos() {
		PhotoManager.CAMERA_SPOTS.forEach(spotPath => {
			const spot = Spot[spotPath];
			const imageProvider: ImageProvider =
				spot.isOfTypeName("DisplaySpot") as DisplaySpot ||
				spot.isOfTypeName("MobileSpot") as MobileSpot;

			if (imageProvider) {
				imageProvider.subscribe(
					'image',
					(
						sender: DisplaySpot | MobileSpot,
						message: {
							readonly filePath: string,	// Path to file just received (typically "/temp/xxx/xxx.jpeg")
							readonly rollName: string	// Camera Block's assigned "roll name"
						}
					) => {
						this.listOfImages.acceptImage(message.filePath);
					}
				)
			}

			imageProvider.subscribe('finish', () => this.listenForPhotos());
		})
	}

	/**
	 * A simple function to be called from a task in order reset everything to a virgin state.
	 */
	@callable('Clear all photos')
	clear() {
		return this.listOfImages.clear(true);
	}
}
